"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getColumnCandidatesForSimpleSelectStatement = exports.getColumnCandidatesForSelectTableExpression = exports.getColumnsForSelectTableExpression = void 0;
/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
const langium_1 = require("langium");
const ast_1 = require("./generated/ast");
function getColumnsForSelectTableExpression(selectTableExpression, onlyAliases = false) {
    if ((0, ast_1.isBinaryTableExpression)(selectTableExpression)) {
        const lhs = getColumnsForSelectTableExpression(selectTableExpression.left, onlyAliases);
        //const rhs = getColumnsForSelectTableExpression(selectTableExpression.right, onlyAliases);
        return lhs;
    }
    else if ((0, ast_1.isParenthesesSelectTableExpression)(selectTableExpression)) {
        return getColumnsForSelectTableExpression(selectTableExpression.value);
    }
    else if ((0, ast_1.isSimpleSelectTableExpression)(selectTableExpression)) {
        return getColumnsSimpleSelectStatement(selectTableExpression.select, onlyAliases);
    }
    else {
        (0, langium_1.assertUnreachable)(selectTableExpression);
    }
    return [];
}
exports.getColumnsForSelectTableExpression = getColumnsForSelectTableExpression;
function getColumnsSimpleSelectStatement(simpleSelectStatement, onlyAliases) {
    return simpleSelectStatement.select.elements.flatMap(e => {
        var _a;
        if ((0, ast_1.isAllStar)(e)) {
            if (onlyAliases) {
                return [];
            }
            const fromAllSources = getColumnCandidatesForSimpleSelectStatement(simpleSelectStatement);
            return fromAllSources.flatMap(t => t);
        }
        else if ((0, ast_1.isAllTable)(e)) {
            if (onlyAliases) {
                return [];
            }
            if (!simpleSelectStatement.from) {
                return [];
            }
            const ref = e.variableName.ref;
            if ((0, ast_1.isTableSourceItem)(ref)) {
                const tableLike = ref.tableName.ref;
                if ((0, ast_1.isTableDefinition)(tableLike)) {
                    const columns = (_a = tableLike.columns.filter(ast_1.isColumnDefinition)) !== null && _a !== void 0 ? _a : [];
                    return columns.map(c => ({
                        name: c.name,
                        typedNode: c.dataType,
                        node: c,
                        isScopedByVariable: true
                    }));
                }
                else if ((0, ast_1.isCommonTableExpression)(tableLike)) {
                    const columns = getColumnsForSelectTableExpression(tableLike.statement);
                    if (tableLike.columnNames.length > 0) {
                        return columns.map((c, i) => (Object.assign(Object.assign({}, c), { name: tableLike.columnNames[i].name })));
                    }
                    return columns;
                }
                else {
                    (0, langium_1.assertUnreachable)(tableLike);
                }
            }
            else if ((0, ast_1.isSubQuerySourceItem)(ref)) {
                return getColumnsForSelectTableExpression(ref.subQuery);
            }
            else {
                (0, langium_1.assertUnreachable)(ref);
            }
        }
        else if ((0, ast_1.isExpressionQuery)(e)) {
            if (e.name) {
                return [{
                        name: e.name,
                        typedNode: e.expr,
                        node: e,
                        isScopedByVariable: false
                    }];
            }
            else {
                if (onlyAliases) {
                    return [];
                }
                const expr = e.expr;
                if ((0, ast_1.isTableRelatedColumnExpression)(expr)) {
                    return resolveColumnNameTypedNode(e, expr.columnName);
                }
                else if ((0, ast_1.isFunctionCall)(expr)) {
                    const functionLike = expr.functionName.ref;
                    return [{
                            name: expr.functionName.$refText,
                            isScopedByVariable: false,
                            node: e,
                            typedNode: functionLike.returnType
                        }];
                }
                else if ((0, ast_1.isColumnNameExpression)(expr)) {
                    const fromAllSources = getColumnCandidatesForSimpleSelectStatement(simpleSelectStatement);
                    const name = expr.columnName.$refText;
                    const column = fromAllSources.find(s => !s.isScopedByVariable && s.name === name);
                    if (column) {
                        return [column];
                    }
                    else {
                        return resolveColumnNameTypedNode(expr, expr.columnName);
                    }
                }
                else if ((0, ast_1.isSubQueryExpression)(expr)) {
                    const columns = getColumnsForSelectTableExpression(expr.subQuery);
                    return [columns[0]];
                }
                else {
                    return [{
                            node: e,
                            typedNode: e.expr,
                            isScopedByVariable: false
                        }];
                }
            }
        }
        else {
            (0, langium_1.assertUnreachable)(e);
        }
        return [];
    });
}
function resolveColumnNameTypedNode(expression, columnName) {
    const ref = columnName.ref;
    let typed = expression;
    if ((0, ast_1.isExpressionQuery)(ref)) {
        typed = ref.expr;
    }
    else if ((0, ast_1.isColumnDefinition)(ref)) {
        typed = ref.dataType;
    }
    return [{
            name: ref === null || ref === void 0 ? void 0 : ref.name,
            isScopedByVariable: false,
            node: expression,
            typedNode: typed
        }];
}
function getColumnCandidatesForSelectTableExpression(selectTableExpression) {
    if ((0, ast_1.isBinaryTableExpression)(selectTableExpression)) {
        const lhs = getColumnCandidatesForSelectTableExpression(selectTableExpression.left);
        //const rhs = getColumnCandidatesForSelectTableExpression(selectTableExpression.right);
        return lhs;
    }
    else if ((0, ast_1.isParenthesesSelectTableExpression)(selectTableExpression)) {
        return getColumnCandidatesForSelectTableExpression(selectTableExpression.value);
    }
    else if ((0, ast_1.isSimpleSelectTableExpression)(selectTableExpression)) {
        return getColumnCandidatesForSimpleSelectStatement(selectTableExpression.select);
    }
    else {
        (0, langium_1.assertUnreachable)(selectTableExpression);
    }
    return [];
}
exports.getColumnCandidatesForSelectTableExpression = getColumnCandidatesForSelectTableExpression;
function getColumnCandidatesForSimpleSelectStatement(selectStatement) {
    var _a, _b;
    const selectElementColumns = getColumnsSimpleSelectStatement(selectStatement, true);
    const fromComputedColumns = (_b = (_a = selectStatement.from) === null || _a === void 0 ? void 0 : _a.sources.list.flatMap(getColumnsForTableSource)) !== null && _b !== void 0 ? _b : [];
    return selectElementColumns.concat(fromComputedColumns);
}
exports.getColumnCandidatesForSimpleSelectStatement = getColumnCandidatesForSimpleSelectStatement;
function getColumnsForTableSource(source) {
    const items = [source.item].concat(source.joins.map(j => j.nextItem));
    return items.flatMap(item => {
        if ((0, ast_1.isTableSourceItem)(item)) {
            const tableLike = item.tableName.ref;
            if ((0, ast_1.isTableDefinition)(tableLike)) {
                return tableLike.columns.filter(ast_1.isColumnDefinition).map(column => ({
                    name: column.name,
                    typedNode: column.dataType,
                    node: column,
                    isScopedByVariable: item.name != null
                }));
            }
            else if ((0, ast_1.isCommonTableExpression)(tableLike)) {
                let columns = getColumnsForSelectTableExpression(tableLike.statement);
                if (tableLike.columnNames.length > 0) {
                    columns = columns.map((c, i) => (Object.assign(Object.assign({}, c), { name: tableLike.columnNames[i].name })));
                }
                return columns;
            }
            return [];
        }
        else if ((0, ast_1.isSubQuerySourceItem)(item)) {
            return getColumnsForSelectTableExpression(item.subQuery);
        }
        else {
            (0, langium_1.assertUnreachable)(item);
            return [];
        }
    });
}
//# sourceMappingURL=sql-type-utilities.js.map