"use strict";
/******************************************************************************
 * Copyright 2022-2023 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
Object.defineProperty(exports, "__esModule", { value: true });
exports.SqlValidator = exports.SqlValidationRegistry = void 0;
const tslib_1 = require("tslib");
const langium_1 = require("langium");
const ast = tslib_1.__importStar(require("./generated/ast"));
const lodash_1 = tslib_1.__importDefault(require("lodash"));
const sql_error_codes_1 = require("./sql-error-codes");
const sql_type_computation_1 = require("./sql-type-computation");
const sql_type_descriptors_1 = require("./sql-type-descriptors");
const sql_type_utilities_1 = require("./sql-type-utilities");
const sql_type_conversion_1 = require("./sql-type-conversion");
class SqlValidationRegistry extends langium_1.ValidationRegistry {
    constructor(services) {
        super(services);
        const validator = services.validation.SqlValidator;
        const checks = {
            TableDefinition: [validator.checkIfTableDefinitionHasAtLeastOneColumn],
            SimpleSelectStatement: [
                validator.checkVariableNamesAreUnique,
                validator.checkIfSelectStatementWithAllStarSelectItemHasAtLeastOneTableSource
            ],
            //Expression: [validator.checkIfExpressionHasType], //TODO uncomment when type system is bullet-proof
            BinaryTableExpression: [validator.checkBinaryTableExpressionMatches],
            IntegerLiteral: [validator.checkIntegerLiteralIsWholeNumber],
            BinaryExpression: [validator.checkBinaryExpressionType],
            UnaryExpression: [validator.checkUnaryExpressionType],
            WhereClause: [validator.checkWhereIsBoolean],
            HavingClause: [validator.checkHavingIsBoolean],
            SubQueryExpression: [validator.checkIfSubQuerySelectsExactlyOneValue]
        };
        this.register(checks, validator);
    }
}
exports.SqlValidationRegistry = SqlValidationRegistry;
class SqlValidator {
    checkBinaryTableExpressionMatches(expr, accept) {
        const lhs = (0, sql_type_utilities_1.getColumnsForSelectTableExpression)(expr.left);
        const rhs = (0, sql_type_utilities_1.getColumnsForSelectTableExpression)(expr.right);
        if (lhs.length !== rhs.length) {
            sql_error_codes_1.ReportAs.TableOperationUsesTablesWithDifferentColumnCounts(expr, {}, accept);
        }
        else {
            lhs.forEach((left, index) => {
                const right = rhs[index];
                const leftType = (0, sql_type_computation_1.computeType)(left.typedNode);
                const rightType = (0, sql_type_computation_1.computeType)(right.typedNode);
                if (leftType && rightType) {
                    if (!(0, sql_type_conversion_1.canConvert)(leftType, rightType) && !(0, sql_type_conversion_1.canConvert)(rightType, leftType)) {
                        sql_error_codes_1.ReportAs.TableOperationUsesTablesWithDifferentColumnTypes(expr, { columnIndex: index }, accept);
                    }
                } //else should be handled by different validator
            });
        }
    }
    checkIfExpressionHasType(expr, accept) {
        const type = (0, sql_type_computation_1.computeType)(expr);
        if (!type) {
            sql_error_codes_1.ReportAs.CannotDeriveTypeOfExpression(expr, {}, accept);
        }
    }
    checkWhereIsBoolean(clause, accept) {
        const type = (0, sql_type_computation_1.computeType)(clause.rowCondition);
        if (type && !(0, sql_type_descriptors_1.isTypeABoolean)(type)) {
            sql_error_codes_1.ReportAs.ExpressionMustReturnABoolean(clause.rowCondition, type, accept);
        }
    }
    checkHavingIsBoolean(clause, accept) {
        const type = (0, sql_type_computation_1.computeType)(clause.groupCondition);
        if (type && !(0, sql_type_descriptors_1.isTypeABoolean)(type)) {
            sql_error_codes_1.ReportAs.ExpressionMustReturnABoolean(clause.groupCondition, type, accept);
        }
    }
    checkBinaryExpressionType(expr, accept) {
        const left = (0, sql_type_computation_1.computeType)(expr.left);
        const right = (0, sql_type_computation_1.computeType)(expr.right);
        const returnType = (0, sql_type_computation_1.computeType)(expr);
        if (left && right && !returnType) {
            sql_error_codes_1.ReportAs.BinaryOperatorNotDefinedForGivenExpressions(expr, {
                left,
                right,
                op: expr.operator
            }, accept);
        }
    }
    checkUnaryExpressionType(expr, accept) {
        const operand = (0, sql_type_computation_1.computeType)(expr.value);
        const returnType = (0, sql_type_computation_1.computeType)(expr);
        if (operand && !returnType) {
            sql_error_codes_1.ReportAs.UnaryOperatorNotDefinedForGivenExpression(expr, {
                operand,
                op: expr.operator
            }, accept);
        }
    }
    checkVariableNamesAreUnique(query, accept) {
        var _a;
        const groups = lodash_1.default.groupBy((_a = query.from) === null || _a === void 0 ? void 0 : _a.sources.list, (s) => s.item.name);
        for (const [key, group] of Object.entries(groups).filter((g) => g[0] && g[1].length > 1)) {
            for (const member of group) {
                sql_error_codes_1.ReportAs.DuplicatedVariableName(member.item, { name: key }, accept);
            }
        }
    }
    checkIntegerLiteralIsWholeNumber(literal, accept) {
        if (Math.floor(literal.value) !== literal.value) {
            sql_error_codes_1.ReportAs.NumericValueIsNotInteger(literal, { value: literal.value }, accept);
        }
    }
    checkIfSelectStatementWithAllStarSelectItemHasAtLeastOneTableSource(selectStatement, accept) {
        if (selectStatement.select.elements.filter(ast.isAllStar).length > 0) {
            if (!selectStatement.from) {
                sql_error_codes_1.ReportAs.AllStarSelectionRequiresTableSources(selectStatement, {}, accept);
            }
        }
    }
    checkIfTableDefinitionHasAtLeastOneColumn(tableDefinition, accept) {
        if (tableDefinition.columns.length === 0) {
            sql_error_codes_1.ReportAs.TableDefinitionRequiresAtLeastOneColumn(tableDefinition, {}, accept);
        }
    }
    //TODO does not hold for insertions! INSERT INTO employees SELECT id, name FROM xxx
    checkIfSubQuerySelectsExactlyOneValue(subQueryExpression, accept) {
        const type = (0, sql_type_computation_1.computeTypeOfSelectStatement)(subQueryExpression.subQuery);
        if (type.discriminator === 'row' && type.columnTypes.length > 1) {
            sql_error_codes_1.ReportAs.SubQueriesWithinSelectStatementsMustHaveExactlyOneColumn(subQueryExpression, {}, accept);
        }
    }
}
exports.SqlValidator = SqlValidator;
//# sourceMappingURL=sql-validator.js.map